@@ -44,6 +44,11 @@ struct cursorObject {
     int notuples:1;          /* 1 if the command was not a SELECT query */
     int withhold:1;          /* 1 if the cursor is named and uses WITH HOLD */
 
+    int scrollable;          /* 1 if the cursor is named and SCROLLABLE,
+                                0 if not scrollable
+                               -1 if undefined (PG may decide scrollable or not)
+                              */
+
     long int rowcount;       /* number of rows affected by last execute */
     long int columns;        /* number of columns fetched from the db */
     long int arraysize;      /* how many rows should fetchmany() return */
@@ -75,6 +80,7 @@ struct cursorObject {
     char *qattr;          /* quoting attr, used when quoting strings */
     char *notice;         /* a notice from the backend */
     char *name;           /* this cursor name */
+    char *qname;          /* this cursor name, quoted */
 
     PyObject *string_types;   /* a set of typecasters for string types */
     PyObject *binary_types;   /* a set of typecasters for binary types */
@@ -84,36 +90,53 @@ struct cursorObject {
 };
 
 
-/* C-callable functions in cursor_int.c and cursor_ext.c */
+/* C-callable functions in cursor_int.c and cursor_type.c */
 BORROWED HIDDEN PyObject *curs_get_cast(cursorObject *self, PyObject *oid);
 HIDDEN void curs_reset(cursorObject *self);
+HIDDEN int psyco_curs_withhold_set(cursorObject *self, PyObject *pyvalue);
+HIDDEN int psyco_curs_scrollable_set(cursorObject *self, PyObject *pyvalue);
+HIDDEN PyObject *psyco_curs_validate_sql_basic(cursorObject *self, PyObject *sql);
 
 /* exception-raising macros */
 #define EXC_IF_CURS_CLOSED(self) \
-if ((self)->closed || ((self)->conn && (self)->conn->closed)) { \
-    PyErr_SetString(InterfaceError, "cursor already closed");   \
-    return NULL; }
+do { \
+    if (!(self)->conn) { \
+        PyErr_SetString(InterfaceError, "the cursor has no connection"); \
+        return NULL; } \
+    if ((self)->closed || (self)->conn->closed) { \
+        PyErr_SetString(InterfaceError, "cursor already closed"); \
+        return NULL; } \
+} while (0)
 
 #define EXC_IF_NO_TUPLES(self) \
-if ((self)->notuples && (self)->name == NULL) {               \
-    PyErr_SetString(ProgrammingError, "no results to fetch"); \
-    return NULL; }
+do \
+    if ((self)->notuples && (self)->name == NULL) { \
+        PyErr_SetString(ProgrammingError, "no results to fetch"); \
+        return NULL; } \
+while (0)
 
 #define EXC_IF_NO_MARK(self) \
-if ((self)->mark != (self)->conn->mark && (self)->withhold == 0) {                                  \
-    PyErr_SetString(ProgrammingError, "named cursor isn't valid anymore"); \
-    return NULL; }
+do \
+    if ((self)->mark != (self)->conn->mark && (self)->withhold == 0) { \
+        PyErr_SetString(ProgrammingError, "named cursor isn't valid anymore"); \
+        return NULL; } \
+while (0)
 
-#define EXC_IF_CURS_ASYNC(self, cmd) if ((self)->conn->async == 1) { \
-    PyErr_SetString(ProgrammingError, #cmd " cannot be used "        \
-    "in asynchronous mode");                                         \
-    return NULL; }
+#define EXC_IF_CURS_ASYNC(self, cmd) \
+do \
+    if ((self)->conn->async == 1) { \
+        PyErr_SetString(ProgrammingError, \
+            #cmd " cannot be used in asynchronous mode"); \
+        return NULL; } \
+while (0)
 
 #define EXC_IF_ASYNC_IN_PROGRESS(self, cmd) \
-if ((self)->conn->async_cursor != NULL) {   \
-    PyErr_SetString(ProgrammingError, #cmd " cannot be used "        \
-    "while an asynchronous query is underway");                      \
-    return NULL; }
+do \
+    if ((self)->conn->async_cursor != NULL) { \
+        PyErr_SetString(ProgrammingError, \
+            #cmd " cannot be used while an asynchronous query is underway"); \
+    return NULL; } \
+while (0)
 
 #ifdef __cplusplus
 }
